# frozen_string_literal: true

module Prawn
  # Namespace for different fonts.
  module Fonts
  end
end

require_relative 'font'
require_relative 'fonts/afm'
require_relative 'fonts/ttf'
require_relative 'fonts/dfont'
require_relative 'fonts/otf'
require_relative 'fonts/ttc'
